@@ -8,7 +8,7 @@
 
 #include <asm/types.h>          /* For __uXX types */
 
-#define IP_VS_VERSION_CODE            0x000906
+#define IP_VS_VERSION_CODE            0x000907
 #define NVERSION(version)                       \
 	(version >> 16) & 0xFF,                 \
 	(version >> 8) & 0xFF,                  \
@@ -252,8 +252,14 @@ extern int ip_vs_get_debug_level(void);
 	    if (level <= ip_vs_get_debug_level())	\
 		    printk(KERN_DEBUG "IPVS: " msg);	\
     } while (0)
+#define IP_VS_DBG_RL(msg...)				\
+    do {						\
+	    if (net_ratelimit())			\
+		    printk(KERN_DEBUG "IPVS: " msg);	\
+    } while (0)
 #else	/* NO DEBUGGING at ALL */
 #define IP_VS_DBG(level, msg...)  do {} while (0)
+#define IP_VS_DBG_RL(msg...)  do {} while (0)
 #endif
 
 #define IP_VS_BUG() BUG()
@@ -261,6 +267,11 @@ extern int ip_vs_get_debug_level(void);
 #define IP_VS_INFO(msg...) printk(KERN_INFO "IPVS: " msg)
 #define IP_VS_WARNING(msg...) \
 	printk(KERN_WARNING "IPVS: " msg)
+#define IP_VS_ERR_RL(msg...)				\
+    do {						\
+	    if (net_ratelimit())			\
+		    printk(KERN_ERR "IPVS: " msg);	\
+    } while (0)
 
 #ifdef CONFIG_IP_VS_DEBUG
 #define EnterFunction(level)						\
@@ -539,10 +550,10 @@ struct ip_vs_app
 	int (*done_conn)(struct ip_vs_app *, struct ip_vs_conn *);
 	/* output hook */
 	int (*pkt_out)(struct ip_vs_app *,
-		       struct ip_vs_conn *, struct sk_buff **);
+		       struct ip_vs_conn *, struct sk_buff *);
 	/* input hook */
 	int (*pkt_in)(struct ip_vs_app *,
-		      struct ip_vs_conn *, struct sk_buff **);
+		      struct ip_vs_conn *, struct sk_buff *);
 };
 
 
@@ -663,8 +674,8 @@ extern int register_ip_vs_app(struct ip_vs_app *mapp,
 extern int unregister_ip_vs_app(struct ip_vs_app *mapp);
 extern struct ip_vs_app * ip_vs_bind_app(struct ip_vs_conn *cp);
 extern int ip_vs_unbind_app(struct ip_vs_conn *cp);
-extern int ip_vs_app_pkt_out(struct ip_vs_conn *, struct sk_buff **skb_p);
-extern int ip_vs_app_pkt_in(struct ip_vs_conn *, struct sk_buff **skb_p);
+extern int ip_vs_app_pkt_out(struct ip_vs_conn *, struct sk_buff *skb);
+extern int ip_vs_app_pkt_in(struct ip_vs_conn *, struct sk_buff *skb);
 extern int ip_vs_skb_replace(struct sk_buff *skb, int pri,
 			     char *o_buf, int o_len, char *n_buf, int n_len);
 extern int ip_vs_app_init(void);
@@ -866,9 +877,9 @@ __ip_vs_get_out_rt(struct ip_vs_conn *cp, u32 rtos)
 			 __ip_vs_dst_check(dest, rtos, 0))) {
 			if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
 				spin_unlock(&dest->dst_lock);
-				IP_VS_DBG(0, "ip_route_output error, "
-					  "dest: %u.%u.%u.%u\n",
-					  NIPQUAD(dest->addr));
+				IP_VS_DBG_RL("ip_route_output error, "
+					     "dest: %u.%u.%u.%u\n",
+					     NIPQUAD(dest->addr));
 				return NULL;
 			}
 			__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
@@ -879,8 +890,8 @@ __ip_vs_get_out_rt(struct ip_vs_conn *cp, u32 rtos)
 		spin_unlock(&dest->dst_lock);
 	} else {
 		if (ip_route_output(&rt, cp->daddr, 0, rtos, 0)) {
-			IP_VS_DBG(0, "ip_route_output error, "
-				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			IP_VS_DBG_RL("ip_route_output error, dest: "
+				     "%u.%u.%u.%u\n", NIPQUAD(dest->addr));
 			return NULL;
 		}
 	}
@@ -911,29 +922,18 @@ static inline void ip_vs_fast_check_update(union ip_vs_tphdr *h,
 		*checkp = 0xFFFF;
 }
 
-static inline struct sk_buff *
-ip_vs_skb_cow(struct sk_buff **skb_p, struct iphdr **iph_p,
-	      unsigned char **t_p)
+static inline int
+ip_vs_skb_cow(struct sk_buff *skb, struct iphdr **iph_p, unsigned char **t_p)
 {
-	struct sk_buff *skb = *skb_p;
-
 	if (skb_cloned(skb)) {
-		struct iphdr *iph;
+		if (pskb_expand_head(skb, 0, 0, GFP_ATOMIC))
+			return -ENOMEM;
 
-		skb = pskb_copy(skb, GFP_ATOMIC);
-		if (!skb)
-			return NULL;
-		/*
-		 *	skb changed, update other pointers
-		 */
-		kfree_skb(*skb_p);
-		*skb_p = skb;
-		iph = skb->nh.iph;
-		*iph_p = iph;
-		*t_p = (char*) iph + iph->ihl * 4;
+		/* skb data changed, update pointers */
+		*iph_p = skb->nh.iph;
+		*t_p = (char*) (*iph_p) + (*iph_p)->ihl * 4;
 	}
-
-	return skb;
+	return 0;
 }
 
 #endif /* __KERNEL__ */
