@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.22 2001/11/04 08:58:43 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.23 2001/11/23 14:34:10 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -610,8 +610,8 @@ static int ip_vs_bypass_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	EnterFunction(10);
 
 	if (ip_route_output(&rt, iph->daddr, 0, RT_TOS(tos), 0)) {
-		IP_VS_DBG(0, "ip_vs_bypass_xmit(): ip_route_output error, "
-			  "dest: %u.%u.%u.%u\n", NIPQUAD(iph->daddr));
+		IP_VS_DBG_RL("ip_vs_bypass_xmit(): ip_route_output error, "
+			     "dest: %u.%u.%u.%u\n", NIPQUAD(iph->daddr));
 		goto tx_error_icmp;
 	}
 
@@ -620,7 +620,7 @@ static int ip_vs_bypass_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
 		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
 		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_bypass_xmit(): frag needed\n");
+		IP_VS_DBG_RL("ip_vs_bypass_xmit(): frag needed\n");
 		goto tx_error;
 	}
 
@@ -696,11 +696,11 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 			if (csum_tcpudp_magic(iph->saddr, iph->daddr,
 					      size, iph->protocol, skb->csum))
 			{
-				IP_VS_DBG(0, "Incoming failed %s checksum "
-					  "from %d.%d.%d.%d (size=%d)!\n",
-					  ip_vs_proto_name(iph->protocol),
-					  NIPQUAD(iph->saddr),
-					  size);
+				IP_VS_DBG_RL("Incoming failed %s checksum "
+					     "from %d.%d.%d.%d (size=%d)!\n",
+					     ip_vs_proto_name(iph->protocol),
+					     NIPQUAD(iph->saddr),
+					     size);
 				goto tx_error;
 			}
 		default:/* CHECKSUM_UNNECESSARY */
@@ -729,7 +729,7 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
 		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
 		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_nat_xmit(): frag needed\n");
+		IP_VS_DBG_RL("ip_vs_nat_xmit(): frag needed\n");
 		goto tx_error;
 	}
 
@@ -738,7 +738,7 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	skb->dst = &rt->u.dst;
 
 	/* copy-on-write the packet before mangling it */
-	if (ip_vs_skb_cow(&skb, &iph, &h.raw) == NULL)
+	if (ip_vs_skb_cow(skb, &iph, &h.raw))
 		return NF_DROP;
 
 	/* mangle the packet */
@@ -749,8 +749,8 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	 *	Attempt ip_vs_app call.
 	 *	will fix ip_vs_conn and iph ack_seq stuff
 	 */
-	if (ip_vs_app_pkt_in(cp, &skb) != 0) {
-		/* skb has probably changed, update pointers. */
+	if (ip_vs_app_pkt_in(cp, skb) != 0) {
+		/* skb data has probably changed, update pointers */
 		iph = skb->nh.iph;
 		h.raw = (char*) iph + ihl;
 		size = skb->len - ihl;
@@ -760,7 +760,7 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	 *	Adjust TCP/UDP checksums
 	 */
 	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		/* Only port and addr are changed, make fast csum update */
+		/* Only port and addr are changed, do fast csum update */
 		ip_vs_fast_check_update(&h, cp->vaddr, cp->daddr,
 					cp->vport, cp->dport, iph->protocol);
 		if (skb->ip_summed == CHECKSUM_HW)
@@ -842,9 +842,9 @@ static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	EnterFunction(10);
 
 	if (skb->protocol != __constant_htons(ETH_P_IP)) {
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): protocol error, "
-			  "ETH_P_IP: %d, skb protocol: %d\n",
-			  __constant_htons(ETH_P_IP), skb->protocol);
+		IP_VS_DBG_RL("ip_vs_tunnel_xmit(): protocol error, "
+			     "ETH_P_IP: %d, skb protocol: %d\n",
+			     __constant_htons(ETH_P_IP), skb->protocol);
 		goto tx_error;
 	}
 
@@ -856,7 +856,7 @@ static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
 	if (mtu < 68) {
 		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): mtu less than 68\n");
+		IP_VS_DBG_RL("ip_vs_tunnel_xmit(): mtu less than 68\n");
 		goto tx_error;
 	}
 	if (skb->dst && mtu < skb->dst->pmtu)
@@ -868,10 +868,13 @@ static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	    && mtu < ntohs(old_iph->tot_len)) {
 		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
 		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): frag needed\n");
+		IP_VS_DBG_RL("ip_vs_tunnel_xmit(): frag needed\n");
 		goto tx_error;
 	}
 
+	if (skb_is_nonlinear(skb))
+		ip_send_check(old_iph);
+
 	skb->h.raw = skb->nh.raw;
 
 	/*
@@ -886,7 +889,7 @@ static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 		if (!new_skb) {
 			ip_rt_put(rt);
 			kfree_skb(skb);
-			IP_VS_ERR("ip_vs_tunnel_xmit(): no memory\n");
+			IP_VS_ERR_RL("ip_vs_tunnel_xmit(): no memory\n");
 			return -EINVAL;
 		}
 		kfree_skb(skb);
@@ -951,13 +954,16 @@ static int ip_vs_dr_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 
 	/* MTU checking */
 	mtu = rt->u.dst.pmtu;
-	if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+	if ((iph->frag_off&__constant_htons(IP_DF)) && skb->len > mtu) {
 		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
 		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_dr_xmit(): frag needed\n");
+		IP_VS_DBG_RL("ip_vs_dr_xmit(): frag needed\n");
 		goto tx_error;
 	}
 
+	if (skb_is_nonlinear(skb) && skb->len <= mtu)
+		ip_send_check(iph);
+
 	/* drop old route */
 	dst_release(skb->dst);
 	skb->dst = &rt->u.dst;
@@ -1219,8 +1225,7 @@ ip_vs_conn_new(int proto, __u32 caddr, __u16 cport, __u32 vaddr, __u16 vport,
 
 	cp = kmem_cache_alloc(ip_vs_conn_cachep, GFP_ATOMIC);
 	if (cp == NULL) {
-		if (net_ratelimit())
-			IP_VS_ERR("ip_vs_conn_new: no memory available.\n");
+		IP_VS_ERR_RL("ip_vs_conn_new: no memory available.\n");
 		return NULL;
 	}
 
