@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.24 2001/11/04 08:58:43 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.25 2001/11/23 14:34:10 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -226,10 +226,8 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 						    dest->addr, 0,
 						    0,
 						    dest);
-			if (ct == NULL) {
-				IP_VS_ERR("ip_vs_conn_new template failed\n");
+			if (ct == NULL)
 				return NULL;
-			}
 
 			ct->timeout = svc->timeout;
 		} else {
@@ -282,10 +280,8 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 						    dest->addr, 0,
 						    0,
 						    dest);
-			if (ct == NULL) {
-				IP_VS_ERR("ip_vs_conn_new template failed\n");
+			if (ct == NULL)
 				return NULL;
-			}
 
 			ct->timeout = svc->timeout;
 		} else {
@@ -305,7 +301,6 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 			    0,
 			    dest);
 	if (cp == NULL) {
-		IP_VS_ERR("ip_vs_conn_new failed\n");
 		ip_vs_conn_put(ct);
 		return NULL;
 	}
@@ -372,11 +367,8 @@ ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 			    dest->addr, dest->port?dest->port:portp[1],
 			    0,
 			    dest);
-	if (cp == NULL) {
-		if (net_ratelimit())
-			IP_VS_ERR("Schedule: ip_vs_conn_new failed\n");
+	if (cp == NULL)
 		return NULL;
-	}
 
 	/*
 	 *    Increase the inactive connection counter because it is in
@@ -425,7 +417,6 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 				    IP_VS_CONN_F_BYPASS,
 				    NULL);
 		if (cp == NULL) {
-			IP_VS_ERR("ip_vs_conn_new failed\n");
 			kfree_skb(skb);
 			return NF_STOLEN;
 		}
@@ -509,25 +500,24 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	struct icmphdr	*icmph;
 	struct iphdr	*ciph;	/* The ip header contained within the ICMP */
 	__u16		*pptr;	/* port numbers from TCP/UDP contained header */
-	int		need_check;
 	unsigned short	ihl;
 	unsigned short	len;
 	unsigned short	clen, csize;
 	struct ip_vs_conn *cp;
 
 	/* reassemble IP fragments, but will it happen in ICMP packets?? */
-	need_check = 0;
 	if (skb->nh.iph->frag_off & htons(IP_MF|IP_OFFSET)) {
 		skb = ip_defrag(skb);
-		if (!skb) {
+		if (!skb)
 			return NF_STOLEN;
-		}
 		*skb_p = skb;
-		need_check = 1;
 	}
 
-	if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0)
-		return NF_DROP;
+	if (skb_is_nonlinear(skb)) {
+		if (skb_linearize(skb, GFP_ATOMIC) != 0)
+			return NF_DROP;
+		ip_send_check(skb->nh.iph);
+	}
 
 	iph = skb->nh.iph;
 	ihl = iph->ihl << 2;
@@ -549,11 +539,8 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	 */
 	if ((icmph->type != ICMP_DEST_UNREACH) &&
 	    (icmph->type != ICMP_SOURCE_QUENCH) &&
-	    (icmph->type != ICMP_TIME_EXCEEDED)) {
-		if (need_check)
-			ip_send_check(iph);
+	    (icmph->type != ICMP_TIME_EXCEEDED))
 		return NF_ACCEPT;
-	}
 
 	/* Now find the contained IP header */
 	clen = len - sizeof(struct icmphdr);
@@ -565,11 +552,8 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 
 	/* We are only interested ICMPs generated from TCP or UDP packets */
-	if (ciph->protocol != IPPROTO_UDP && ciph->protocol != IPPROTO_TCP) {
-		if (need_check)
-			ip_send_check(iph);
+	if (ciph->protocol != IPPROTO_UDP && ciph->protocol != IPPROTO_TCP)
 		return NF_ACCEPT;
-	}
 
 	/* We need at least TCP/UDP ports here */
 	if (clen < csize + sizeof(struct udphdr))
@@ -598,11 +582,8 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	/* ciph content is actually <protocol, caddr, cport, daddr, dport> */
 	cp = ip_vs_conn_out_get(ciph->protocol, ciph->daddr, pptr[1],
 				ciph->saddr, pptr[0]);
-	if (!cp) {
-		if (need_check)
-			ip_send_check(iph);
+	if (!cp)
 		return NF_ACCEPT;
-	}
 
 	if (IP_VS_FWD_METHOD(cp) != 0) {
 		IP_VS_ERR("shouldn't reach here, because the box is on the"
@@ -658,7 +639,6 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	struct ip_vs_conn *cp;
 	int size;
 	int ihl;
-	int need_check;
 
 	EnterFunction(11);
 
@@ -674,15 +654,12 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		return NF_ACCEPT;
 
 	/* reassemble IP fragments */
-	need_check = 0;
 	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
 		skb = ip_defrag(skb);
-		if (!skb) {
+		if (!skb)
 			return NF_STOLEN;
-		}
 		iph = skb->nh.iph;
 		*skb_p = skb;
-		need_check = 1;
 	}
 
 	/* make sure that protocol header available in skb data area,
@@ -718,7 +695,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 			  ip_vs_proto_name(iph->protocol),
 			  NIPQUAD(iph->daddr),
 			  ntohs(h.portp[1]));
-		if (need_check)
+		if (skb_is_nonlinear(skb))
 			ip_send_check(iph);
 		return NF_ACCEPT;
 	}
@@ -753,11 +730,11 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 			if (csum_tcpudp_magic(iph->saddr, iph->daddr, size,
 					      iph->protocol, skb->csum)) {
 				ip_vs_conn_put(cp);
-				IP_VS_DBG(0, "Outgoing failed %s checksum "
-					  "from %d.%d.%d.%d (size=%d)!\n",
-					  ip_vs_proto_name(iph->protocol),
-					  NIPQUAD(iph->saddr),
-					  size);
+				IP_VS_DBG_RL("Outgoing failed %s checksum "
+					     "from %d.%d.%d.%d (size=%d)!\n",
+					     ip_vs_proto_name(iph->protocol),
+					     NIPQUAD(iph->saddr),
+					     size);
 				return NF_DROP;
 			}
 		default:
@@ -777,9 +754,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	/*
 	 *	Call application helper if needed
 	 */
-	if (ip_vs_app_pkt_out(cp, skb_p) != 0) {
-		/* skb has probably changed, update pointers */
-		skb = *skb_p;
+	if (ip_vs_app_pkt_out(cp, skb) != 0) {
+		/* skb data has probably changed, update pointers */
 		iph = skb->nh.iph;
 		h.raw = (char*)iph + ihl;
 		size = skb->len - ihl;
@@ -789,7 +765,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	 *	Adjust TCP/UDP checksums
 	 */
 	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		/* Only port and addr are changed, make fast csum update */
+		/* Only port and addr are changed, do fast csum update */
 		ip_vs_fast_check_update(&h, cp->daddr, cp->vaddr,
 					cp->dport, cp->vport, iph->protocol);
 		if (skb->ip_summed == CHECKSUM_HW)
@@ -879,8 +855,11 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	struct rtable *rt;			/* Route to the other host */
 	int    mtu;
 
-	if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0)
-		return NF_DROP;
+	if (skb_is_nonlinear(skb)) {
+		if (skb_linearize(skb, GFP_ATOMIC) != 0)
+			return NF_DROP;
+		ip_send_check(skb->nh.iph);
+	}
 
 	iph = skb->nh.iph;
 	icmph = (struct icmphdr *)((char *)iph+(iph->ihl<<2));
@@ -910,8 +889,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 
 	/* We are only interested ICMPs generated from TCP or UDP packets */
-	if ((ciph->protocol != IPPROTO_UDP) &&
-	    (ciph->protocol != IPPROTO_TCP))
+	if (ciph->protocol != IPPROTO_UDP && ciph->protocol != IPPROTO_TCP)
 		return NF_ACCEPT;
 
 	/* We need at least TCP/UDP ports here */
@@ -921,8 +899,8 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	/* Ensure the checksum is correct */
 	if (ip_compute_csum((unsigned char *) icmph, len)) {
 		/* Failed checksum! */
-		IP_VS_ERR("incoming ICMP: failed checksum from %d.%d.%d.%d!\n",
-			  NIPQUAD(iph->saddr));
+		IP_VS_ERR_RL("incoming ICMP: failed checksum from "
+			     "%d.%d.%d.%d!\n", NIPQUAD(iph->saddr));
 		return NF_DROP;
 	}
 
@@ -958,8 +936,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	}
 
 	/* copy-on-write the packet before mangling it */
-	if ((skb = ip_vs_skb_cow(skb_p, &iph,
-				 (unsigned char**)&icmph)) == NULL) {
+	if (ip_vs_skb_cow(skb, &iph, (unsigned char**)&icmph)) {
 		ip_vs_conn_put(cp);
 		return NF_DROP;
 	}
@@ -1001,7 +978,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
 		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
 		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_in_icmp(): frag needed\n");
+		IP_VS_DBG_RL("ip_vs_in_icmp(): frag needed\n");
 		goto tx_error;
 	}
 
@@ -1141,7 +1118,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	if (cp->packet_xmit)
 		ret = cp->packet_xmit(skb, cp);
 	else {
-		IP_VS_DBG(0, "warning: packet_xmit is null");
+		IP_VS_DBG_RL("warning: packet_xmit is null");
 		ret = NF_ACCEPT;
 	}
 
@@ -1181,16 +1158,10 @@ static unsigned int ip_vs_forward_icmp(unsigned int hooknum,
 		return NF_ACCEPT;
 
 	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
-		int ret;
-		
 		skb = ip_defrag(skb);
 		if (!skb)
 			return NF_STOLEN;
 		*skb_p = skb;
-		ret = ip_vs_in_icmp(skb_p);
-		if (ret == NF_ACCEPT)
-			ip_send_check((*skb_p)->nh.iph);
-		return ret;
 	}
 
 	return ip_vs_in_icmp(skb_p);
