@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_ctl.c,v 1.21 2001/09/19 01:14:24 wensong Exp $
+ * Version:     $Id: ip_vs_ctl.c,v 1.22 2001/11/04 08:58:43 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -990,12 +990,12 @@ ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 	/*
 	 * Lookup the scheduler, by 'ur->sched_name'
 	 */
-	sched = ip_vs_lookup_scheduler(ur->sched_name);
+	sched = ip_vs_scheduler_get(ur->sched_name);
 	if (sched == NULL) {
 		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
 			   ur->sched_name);
 		ret = -ENOENT;
-		goto out_err;
+		goto out_mod_dec;
 	}
 
 	svc = (struct ip_vs_service*)
@@ -1041,6 +1041,9 @@ ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 	atomic_set(&svc->usecnt, 1);
 	atomic_set(&svc->refcnt, 0);
 
+	ip_vs_new_estimator(&svc->stats);
+	ip_vs_num_services++;
+
 	/*
 	 *    Hash the service into the service table
 	 */
@@ -1048,17 +1051,14 @@ ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
 	ip_vs_svc_hash(svc);
 	write_unlock_bh(&__ip_vs_svc_lock);
 
-	ip_vs_new_estimator(&svc->stats);
-	ip_vs_num_services++;
 	*svc_p = svc;
 	return 0;
 
   out_err:
-	if (sched->module)
-		__MOD_DEC_USE_COUNT(sched->module);
-	if (svc) {
+	if (svc)
 		kfree(svc);
-	}
+	ip_vs_scheduler_put(sched);
+  out_mod_dec:
 	MOD_DEC_USE_COUNT;
 	return ret;
 }
@@ -1076,7 +1076,7 @@ static int ip_vs_edit_service(struct ip_vs_service *svc,
 	/*
 	 * Lookup the scheduler, by 'ur->sched_name'
 	 */
-	sched = ip_vs_lookup_scheduler(ur->sched_name);
+	sched = ip_vs_scheduler_get(ur->sched_name);
 	if (sched == NULL) {
 		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
 			   ur->sched_name);
@@ -1129,8 +1129,8 @@ static int ip_vs_edit_service(struct ip_vs_service *svc,
   out:
 	write_unlock_bh(&__ip_vs_svc_lock);
 
-	if (old_sched && old_sched->module)
-		__MOD_DEC_USE_COUNT(old_sched->module);
+	if (old_sched)
+		ip_vs_scheduler_put(old_sched);
 
 	return ret;
 }
@@ -1260,6 +1260,58 @@ static int ip_vs_flush(void)
 }
 
 
+/*
+ *  Zero counters in a service or all services
+ */
+static inline void
+__ip_vs_zero_stats(struct ip_vs_stats *stats)
+{
+	spin_lock_bh(&stats->lock);
+	memset(stats, 0, (char *)&stats->lock - (char *)stats);
+	spin_unlock_bh(&stats->lock);
+	ip_vs_zero_estimator(stats);
+}
+
+static int ip_vs_zero_service(struct ip_vs_service *svc)
+{
+	struct list_head *l;
+	struct ip_vs_dest *dest;
+
+	write_lock_bh(&__ip_vs_svc_lock);
+	list_for_each (l, &svc->destinations) {
+		dest = list_entry(l, struct ip_vs_dest, n_list);
+		__ip_vs_zero_stats(&dest->stats);
+	}
+	__ip_vs_zero_stats(&svc->stats);
+	write_unlock_bh(&__ip_vs_svc_lock);
+	return 0;
+}
+
+static int ip_vs_zero_all(void)
+{
+	int idx;
+	struct list_head *l;
+	struct ip_vs_service *svc;
+
+	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+		list_for_each (l, &ip_vs_svc_table[idx]) {
+			svc = list_entry(l, struct ip_vs_service, s_list);
+			ip_vs_zero_service(svc);
+		}
+	}
+
+	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
+		list_for_each (l, &ip_vs_svc_fwm_table[idx]) {
+			svc = list_entry(l, struct ip_vs_service, f_list);
+			ip_vs_zero_service(svc);
+		}
+	}
+
+	__ip_vs_zero_stats(&ip_vs_stats);
+	return 0;
+}
+
+
 static int ip_vs_sysctl_defense_mode(ctl_table *ctl, int write,
 	struct file * filp, void *buffer, size_t *lenp)
 {
@@ -1376,6 +1428,7 @@ static struct ip_vs_sysctl_table ipv4_vs_table = {
 
 /*
  *	Write the contents of the VS rule table to a PROCfs file.
+ *	(It is kept just for backward compatibility)
  */
 static inline char *ip_vs_fwd_name(unsigned flags)
 {
@@ -1644,14 +1697,19 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 		ret = ip_vs_set_timeouts(urule);
 		goto out_unlock;
 	} else if (cmd == IP_VS_SO_SET_STARTDAEMON) {
-		ret = start_sync_thread(urule);
+		ret = start_sync_thread(urule->state, urule->mcast_ifn);
 		goto out_unlock;
 	} else if (cmd == IP_VS_SO_SET_STOPDAEMON) {
 		ret = stop_sync_thread();
 		goto out_unlock;
+	} else if (cmd == IP_VS_SO_SET_ZERO) {
+		/* if no service address is set, zero counters in all */
+		if (!urule->vfwmark && !urule->vaddr && !urule->vport) {
+			ret = ip_vs_zero_all();
+			goto out_unlock;
+		}
 	}
 
-
 	/*
 	 * Check for valid protocol: TCP or UDP. Even for fwmark!=0
 	 */
@@ -1664,7 +1722,7 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 	}
 
 	/*
-	 * Lookup the exact service by (protocol, vaddr, vport) or fwmark
+	 * Lookup the exact service by <protocol, vaddr, vport> or fwmark
 	 */
 	if (urule->vfwmark == 0)
 		svc = __ip_vs_service_get(urule->protocol,
@@ -1672,6 +1730,12 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 	else
 		svc = __ip_vs_svc_fwm_get(urule->vfwmark);
 
+	if (cmd != IP_VS_SO_SET_ADD
+	    && (svc == NULL || svc->protocol != urule->protocol)) {
+		ret = -ESRCH;
+		goto out_unlock;
+	}
+
 	switch (cmd) {
 	case IP_VS_SO_SET_ADD:
 		if (svc != NULL)
@@ -1680,37 +1744,24 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 			ret = ip_vs_add_service(urule, &svc);
 		break;
 	case IP_VS_SO_SET_EDIT:
-		if (svc == NULL || svc->protocol != urule->protocol)
-			ret = -ESRCH;
-		else
-			ret = ip_vs_edit_service(svc, urule);
+		ret = ip_vs_edit_service(svc, urule);
 		break;
 	case IP_VS_SO_SET_DEL:
-		if (svc == NULL || svc->protocol != urule->protocol)
-			ret =  -ESRCH;
-		else {
-			ret = ip_vs_del_service(svc);
-			if (!ret)
-				goto out_unlock;
-		}
+		ret = ip_vs_del_service(svc);
+		if (!ret)
+			goto out_unlock;
 		break;
 	case IP_VS_SO_SET_ADDDEST:
-		if (svc == NULL || svc->protocol != urule->protocol)
-			ret = -ESRCH;
-		else
-			ret = ip_vs_add_dest(svc, urule);
+		ret = ip_vs_add_dest(svc, urule);
 		break;
 	case IP_VS_SO_SET_EDITDEST:
-		if (svc == NULL || svc->protocol != urule->protocol)
-			ret = -ESRCH;
-		else
-			ret = ip_vs_edit_dest(svc, urule);
+		ret = ip_vs_edit_dest(svc, urule);
 		break;
 	case IP_VS_SO_SET_DELDEST:
-		if (svc == NULL || svc->protocol != urule->protocol)
-			ret = -ESRCH;
-		else
-			ret = ip_vs_del_dest(svc, urule);
+		ret = ip_vs_del_dest(svc, urule);
+		break;
+	case IP_VS_SO_SET_ZERO:
+		ret = ip_vs_zero_service(svc);
 		break;
 	default:
 		ret = -EINVAL;
@@ -1731,9 +1782,9 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 static inline void
 __ip_vs_copy_stats(struct ip_vs_stats_user *dst, struct ip_vs_stats *src)
 {
-	spin_lock(&src->lock);
+	spin_lock_bh(&src->lock);
 	memcpy(dst, src, (char*)&src->lock - (char*)src);
-	spin_unlock(&src->lock);
+	spin_unlock_bh(&src->lock);
 }
 
 static inline int
@@ -1746,7 +1797,6 @@ __ip_vs_get_service_entries(const struct ip_vs_get_services *get,
 	struct ip_vs_service_user entry;
 	int ret = 0;
 
-	down_interruptible(&__ip_vs_mutex);
 	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
 		list_for_each (l, &ip_vs_svc_table[idx]) {
 			if (count >= get->num_services)
@@ -1795,7 +1845,6 @@ __ip_vs_get_service_entries(const struct ip_vs_get_services *get,
 		}
 	}
  out:
-	up(&__ip_vs_mutex);
 	return ret;
 }
 
@@ -1806,7 +1855,6 @@ __ip_vs_get_dest_entries(const struct ip_vs_get_dests *get,
 	struct ip_vs_service *svc;
 	int ret = 0;
 
-	down_interruptible(&__ip_vs_mutex);
 	if (get->fwmark)
 		svc = __ip_vs_svc_fwm_get(get->fwmark);
 	else
@@ -1840,7 +1888,6 @@ __ip_vs_get_dest_entries(const struct ip_vs_get_dests *get,
 		ip_vs_service_put(svc);
 	} else
 		ret = -ESRCH;
-	up(&__ip_vs_mutex);
 	return ret;
 }
 
@@ -1853,6 +1900,8 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 	if (!capable(CAP_NET_ADMIN))
 		return -EPERM;
 
+	down_interruptible(&__ip_vs_mutex);
+
 	switch (cmd) {
 	case IP_VS_SO_GET_VERSION:
 	{
@@ -1910,7 +1959,6 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 		if (copy_from_user(&get, user, sizeof(get)))
 			return -EFAULT;
 
-		down_interruptible(&__ip_vs_mutex);
 		if (get.fwmark)
 			svc = __ip_vs_svc_fwm_get(get.fwmark);
 		else
@@ -1928,7 +1976,6 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 			ip_vs_service_put(svc);
 		} else
 			ret = -ESRCH;
-		up(&__ip_vs_mutex);
 	}
 	break;
 
@@ -1977,7 +2024,7 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 			return -EINVAL;
 		}
 		u.state = ip_vs_sync_state;
-		strcpy(u.mcast_ifn, mcast_ifn);
+		strcpy(u.mcast_ifn, ip_vs_mcast_ifn);
 		if (copy_to_user(user, &u, sizeof(u)) != 0)
 			ret = -EFAULT;
 	}
@@ -1987,6 +2034,7 @@ do_ip_vs_get_ctl(struct sock *sk, int cmd, void *user, int *len)
 		ret = -EINVAL;
 	}
 
+	up(&__ip_vs_mutex);
 	return ret;
 }
 
