@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_sync.c,v 1.3 2001/09/18 15:42:54 wensong Exp $
+ * Version:     $Id: ip_vs_sync.c,v 1.4 2001/11/04 08:58:43 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -281,10 +281,10 @@ static void ip_vs_process_message(const char *buffer, const size_t buflen)
 
 
 /* ipvs sync daemon state */
-int ip_vs_sync_state;
+int ip_vs_sync_state = IP_VS_STATE_NONE;
 
 /* multicast interface name */
-char mcast_ifn[IP_VS_IFNAME_MAXLEN];
+char ip_vs_mcast_ifn[IP_VS_IFNAME_MAXLEN];
 
 /* multicast addr */
 static struct sockaddr_in mcast_addr;
@@ -363,6 +363,31 @@ join_mcast_group(struct sock *sk, struct in_addr *addr, char *ifname)
 }
 
 
+static int bind_mcastif_addr(struct socket *sock, char *ifname)
+{
+	struct net_device *dev;
+	u32 addr;
+	struct sockaddr_in sin;
+
+	if ((dev = __dev_get_by_name(ifname)) == NULL)
+		return -ENODEV;
+
+	addr = inet_select_addr(dev, 0, RT_SCOPE_UNIVERSE);
+	if (!addr)
+		IP_VS_ERR("You probably need to specify IP address on "
+			  "multicast interface.\n");
+
+	IP_VS_DBG(7, "binding socket with (%s) %u.%u.%u.%u\n",
+		  ifname, NIPQUAD(addr));
+
+	/* Now bind the socket with the address of multicast interface */
+	sin.sin_family	     = AF_INET;
+	sin.sin_addr.s_addr  = addr;
+	sin.sin_port         = 0;
+
+	return sock->ops->bind(sock, (struct sockaddr*)&sin, sizeof(sin));
+}
+
 /*
  *      Set up sending multicast socket over UDP
  */
@@ -376,22 +401,31 @@ static struct socket * make_send_sock(void)
 		return NULL;
 	}
 
-	if (set_mcast_if(sock->sk, mcast_ifn) < 0) {
+	if (set_mcast_if(sock->sk, ip_vs_mcast_ifn) < 0) {
 		IP_VS_ERR("Error setting outbound mcast interface\n");
-		return NULL;
+		goto error;
 	}
 
 	set_mcast_loop(sock->sk, 0);
 	set_mcast_ttl(sock->sk, 1);
 
+	if (bind_mcastif_addr(sock, ip_vs_mcast_ifn) < 0) {
+		IP_VS_ERR("Error binding address of the mcast interface\n");
+		goto error;
+	}
+
 	if (sock->ops->connect(sock,
 			       (struct sockaddr*)&mcast_addr,
 			       sizeof(struct sockaddr), 0) < 0) {
 		IP_VS_ERR("Error connecting to the multicast addr\n");
-		return NULL;
+		goto error;
 	}
 
 	return sock;
+
+  error:
+	sock_release(sock);
+	return NULL;
 }
 
 
@@ -415,18 +449,22 @@ static struct socket * make_receive_sock(void)
 			    (struct sockaddr*)&mcast_addr,
 			    sizeof(struct sockaddr)) < 0) {
 		IP_VS_ERR("Error binding to the multicast addr\n");
-		return NULL;
+		goto error;
 	}
 
 	/* join the multicast group */
 	if (join_mcast_group(sock->sk,
 			     (struct in_addr*)&mcast_addr.sin_addr,
-			     mcast_ifn) < 0) {
+			     ip_vs_mcast_ifn) < 0) {
 		IP_VS_ERR("Error joining to the multicast group\n");
-		return NULL;
+		goto error;
 	}
 
 	return sock;
+
+  error:
+	sock_release(sock);
+	return NULL;
 }
 
 
@@ -439,6 +477,8 @@ ip_vs_send_async(struct socket *sock, const char *buffer, const size_t length)
 	int		len;
 
 	EnterFunction(7);
+	iov.iov_base     = (void *)buffer;
+	iov.iov_len      = length;
 	msg.msg_name     = 0;
 	msg.msg_namelen  = 0;
 	msg.msg_iov	 = &iov;
@@ -446,8 +486,6 @@ ip_vs_send_async(struct socket *sock, const char *buffer, const size_t length)
 	msg.msg_control  = NULL;
 	msg.msg_controllen = 0;
 	msg.msg_flags    = MSG_DONTWAIT|MSG_NOSIGNAL;
-	msg.msg_iov->iov_base = (void *)buffer;
-	msg.msg_iov->iov_len  = length;
 
 	oldfs = get_fs(); set_fs(KERNEL_DS);
 	len = sock_sendmsg(sock, &msg, (size_t)(length));
@@ -469,6 +507,8 @@ ip_vs_receive(struct socket *sock, char *buffer, const size_t buflen)
 	EnterFunction(7);
 
 	/* Receive a packet */
+	iov.iov_base     = buffer;
+	iov.iov_len      = (size_t)buflen;
 	msg.msg_name     = 0;
 	msg.msg_namelen  = 0;
 	msg.msg_iov	 = &iov;
@@ -477,11 +517,8 @@ ip_vs_receive(struct socket *sock, char *buffer, const size_t buflen)
 	msg.msg_controllen = 0;
 	msg.msg_flags    = 0;
 
-	iov.iov_base = buffer;
-	iov.iov_len  = (size_t)buflen-1;
-
 	oldfs = get_fs(); set_fs(KERNEL_DS);
-	len = sock_recvmsg(sock, &msg, buflen-1, 0);
+	len = sock_recvmsg(sock, &msg, buflen, 0);
 	set_fs(oldfs);
 
 	if (len < 0)
@@ -569,7 +606,7 @@ static void sync_backup_loop(void)
 	/* create the receiving multicast socket */
 	sock = make_receive_sock();
 	if (!sock)
-		return;
+		goto out;
 
 	for (;;) {
 		/* do you have data now? */
@@ -591,6 +628,9 @@ static void sync_backup_loop(void)
 
 	/* release the sending multicast socket */
 	sock_release(sock);
+
+  out:
+	kfree(buf);
 }
 
 
@@ -604,6 +644,8 @@ static int sync_thread(void *data)
 
 	oldmm = get_fs();
 	set_fs(KERNEL_DS);
+
+	sync_pid = current->pid;
 	IP_VS_INFO("sync thread started.\n");
 
 	if (ip_vs_sync_state == IP_VS_STATE_MASTER)
@@ -633,6 +675,10 @@ static int sync_thread(void *data)
 
 	remove_wait_queue(&sync_wait, &wait);
 
+	/* thread exits */
+	sync_pid = 0;
+	IP_VS_INFO("sync thread stopped!\n");
+
 	stop_sync = 0;
 	wake_up(&stop_sync_wait);
 
@@ -642,16 +688,15 @@ static int sync_thread(void *data)
 	return 0;
 }
 
-int start_sync_thread(struct ip_vs_rule_user *ur)
+int start_sync_thread(int state, char *mcast_ifn)
 {
 	if (sync_pid)
 		return -EEXIST;
 
-	ip_vs_sync_state = ur->state;
-	strcpy(mcast_ifn, ur->mcast_ifn);
+	ip_vs_sync_state = state;
+	strcpy(ip_vs_mcast_ifn, mcast_ifn);
 
-	sync_pid = kernel_thread(sync_thread, NULL, 0);
-	if (sync_pid < 0)
+	if (kernel_thread(sync_thread, NULL, 0) < 0)
 		IP_VS_BUG();
 	return 0;
 }
@@ -680,10 +725,8 @@ int stop_sync_thread(void)
 	sys_wait4(sync_pid, NULL, __WCLONE, NULL); */
 	waitpid(sync_pid, NULL, __WCLONE);
 
-	sync_pid = 0;
 	if (stop_sync)
 		IP_VS_BUG();
 
-	IP_VS_INFO("sync thread stopped!\n");
 	return 0;
 }
